from django.core.management import BaseCommand
from eveuniverse.models import EveGroup

from fleetcomp.models import ShipGrouping


class Command(BaseCommand):

    help = "Creates default groupings in the database"

    def handle(self, *args, **kwargs):
        """Add default objects"""

        logistics, _ = ShipGrouping.objects.get_or_create(
            column_index=10,
            display_name="Logistics",
        )
        logi_cruisers, _ = EveGroup.objects.get_or_create_esi(id=832)
        logi_frigates, _ = EveGroup.objects.get_or_create_esi(id=1527)
        logistics.associated_groups.add(logi_cruisers, logi_frigates)

        hic, _ = ShipGrouping.objects.get_or_create(
            column_index=10,
            display_name="HIC",
        )
        hic_group, _ = EveGroup.objects.get_or_create_esi(id=894)
        hic.associated_groups.add(hic_group)

        links, _ = ShipGrouping.objects.get_or_create(
            column_index=10,
            display_name="Link Ships",
        )
        command_ships_group, _ = EveGroup.objects.get_or_create_esi(id=540)
        links.associated_groups.add(command_ships_group)

        cynoes, _ = ShipGrouping.objects.get_or_create(
            column_index=10,
            display_name="Cynoes",
        )
        force_recon_group, _ = EveGroup.objects.get_or_create_esi(id=833)
        cynoes.associated_groups.add(force_recon_group)

        dreads, _ = ShipGrouping.objects.get_or_create(
            column_index=20,
            display_name="Dreadnoughts",
        )
        dreadnought_group, _ = EveGroup.objects.get_or_create_esi(id=485)
        dreads.associated_groups.add(dreadnought_group)

        faxes, _ = ShipGrouping.objects.get_or_create(
            column_index=20,
            display_name="FAXes",
        )
        force_auxiliary_group, _ = EveGroup.objects.get_or_create_esi(id=1538)
        faxes.associated_groups.add(force_auxiliary_group)



"""Admin site."""

from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveGroup, EveType

from fleetcomp.models import FleetMember, FleetSnapshot, ShipGrouping

# Register your models for the admin site here.

EVE_SHIP_CATEGORY_ID = 6


class FleetMembersInline(admin.TabularInline):
    model = FleetMember

    def has_add_permission(self, request, obj):
        return False

    def has_change_permission(self, request, obj=...):
        return False

    def has_delete_permission(self, request, obj=...):
        return False


@admin.register(FleetSnapshot)
class FleetSnapshotAdmin(admin.ModelAdmin):
    list_display = ["commander", "timestamp", "fleet_id"]
    inlines = [FleetMembersInline]

    def has_change_permission(self, request, obj=...):
        return False


@admin.register(ShipGrouping)
class ShipGroupingAdmin(admin.ModelAdmin):
    list_display = ["display_name", "column_index", "ship_types", "ship_groups"]
    ordering = ["column_index"]
    filter_horizontal = ("associated_types", "associated_groups")

    fieldsets = (
        (
            None,
            {
                "fields": ("display_name", "column_index"),
            },
        ),
        (
            _("Associated entities"),
            {
                "fields": (
                    "associated_types",
                    "associated_groups",
                )
            },
        ),
    )

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        if db_field.name == "associated_types":
            kwargs["queryset"] = EveType.objects.filter(
                eve_group__eve_category=EVE_SHIP_CATEGORY_ID
            ).order_by("name")
        elif db_field.name == "associated_groups":
            kwargs["queryset"] = EveGroup.objects.filter(
                eve_category=EVE_SHIP_CATEGORY_ID
            ).order_by("name")
        return super().formfield_for_manytomany(db_field, request, **kwargs)

    @admin.display(description=_("Ship types"))
    def ship_types(self, custom_grouping: ShipGrouping):
        return ", ".join(
            custom_grouping.associated_types.values_list("name", flat=True)
        )

    @admin.display(description=_("Ship groups"))
    def ship_groups(self, custom_grouping: ShipGrouping):
        return ", ".join(
            custom_grouping.associated_groups.values_list("name", flat=True)
        )



"""Esi interactions"""

from dataclasses import dataclass

from bravado.exception import HTTPNotFound

from esi.clients import EsiClientProvider

from allianceauth.services.hooks import get_extension_logger

from . import __version__, models

logger = get_extension_logger(__name__)

esi = EsiClientProvider(app_info_text=f"aa-fleetcomp v{__version__}")


class CharacterNotInFleet(Exception):
    """Yield when the ESI returns a 404 error because the character isn't in fleet"""


class NoAccessToFleet(Exception):
    """Yield when the ESI returns a 404 error for a lack of access"""


@dataclass
class FleetData:
    """Fleet information pulled from ESI"""

    fleet_id: int
    fleet_commander: "models.FleetCommander"
    fleet_members: list["FleetMemberData"]


@dataclass
class FleetMemberData:
    """Information of a fleet member from the ESI"""

    character_id: int
    ship_type_id: int

    @classmethod
    def create_from_member_list(
        cls, fleet_members_info: list[dict]
    ) -> list["FleetMemberData"]:
        """Creates the list of member information from the ESI endpoint data"""
        return [
            FleetMemberData(
                character_id=fleet_member_info["character_id"],
                ship_type_id=fleet_member_info["ship_type_id"],
            )
            for fleet_member_info in fleet_members_info
        ]


def get_fleet_id(fleet_commander: "models.FleetCommander") -> int:
    """Return the current fleet id of this fleet commander"""

    fleet_info = __get_character_fleet_info(fleet_commander)
    fleet_id = fleet_info["fleet_id"]

    return fleet_id


def get_fleet_data(fleet_commander: "models.FleetCommander") -> FleetData:
    """
    Returns all data from the fleet FCed by this fleet commander
    """
    logger.info("Trying to query the fleet data of commander %s", fleet_commander)

    fleet_id = get_fleet_id(fleet_commander)
    fleet_members = __get_fleet_members(fleet_commander, fleet_id)

    logger.debug(fleet_commander)

    return FleetData(
        fleet_id=fleet_id,
        fleet_commander=fleet_commander,
        fleet_members=FleetMemberData.create_from_member_list(fleet_members),
    )


def __get_character_fleet_info(fleet_member: "models.FleetCommander"):
    """
    Pulls basic fleet information (fleet id, position in fleet)
    60 sec cache
    """

    try:
        fleet_info = esi.client.Fleets.get_characters_character_id_fleet(
            character_id=fleet_member.character_id,
            token=fleet_member.fetch_token().valid_access_token(),
        ).result()
    except HTTPNotFound as e:
        if "Character is not in a fleet" == e.swagger_result.get("error"):
            raise CharacterNotInFleet(
                "The character doesn't appear to be in a fleet"
            ) from e
        raise

    logger.debug(fleet_info)

    return fleet_info


def __get_fleet_members(fleet_member: "models.FleetCommander", fleet_id: int):
    """
    Pulls the fleet members information from the ESI
    5 sec cache
    """

    try:
        fleet_members = esi.client.Fleets.get_fleets_fleet_id_members(
            fleet_id=fleet_id,
            token=fleet_member.fetch_token().valid_access_token(),
        ).results()
    except HTTPNotFound as e:
        if "Character is not in a fleet" == e.swagger_result.get("error"):
            raise CharacterNotInFleet(
                "The character doesn't appear to be in a fleet"
            ) from e
        if (
            "The fleet does not exist or you don't have access to it!"
            == e.swagger_result.get("error")
        ):
            raise NoAccessToFleet(
                "The character doesn't appear to be a fleet commander"
            ) from e
        raise

    return fleet_members



"""Views."""

from django.contrib import messages
from django.contrib.auth.decorators import permission_required
from django.contrib.auth.models import User
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import cache_page
from esi.decorators import token_required

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger

from fleetcomp import ESI_SCOPES
from fleetcomp.app_settings import FLEETCOMP_FLEET_CACHE_MINUTES
from fleetcomp.esi import (
    CharacterNotInFleet,
    NoAccessToFleet,
    get_fleet_data,
    get_fleet_id,
)
from fleetcomp.models import FleetCommander, FleetSnapshot, ShipGrouping

logger = get_extension_logger(__name__)


@permission_required("fleetcomp.basic_access")
def index(request):
    """Render index view."""
    snapshots = FleetSnapshot.objects.all()
    if not request.user.has_perm("fleetcomp.view_all"):
        snapshots.filter(commander__character_ownership__user=request.user)
    return render(request, "fleetcomp/index.html", {"snapshots": snapshots})


@permission_required("fleetcomp.basic_access")
@token_required(scopes=ESI_SCOPES)
def capture_own_fleet_composition(request, token):
    """
    Adds a new fleet commander to the list of fleet commanders this user can pull data from.
    """
    logger.info("Capturing fleet composition from token %s", token)

    character_ownership: CharacterOwnership = get_object_or_404(
        request.user.character_ownerships.select_related("character"),
        character__character_id=token.character_id,
    )

    fleet_commander, created = FleetCommander.objects.get_or_create(
        character_ownership=character_ownership,
    )

    if created:
        logger.info(
            "Created a new fleet commander from char_ownership: %s", character_ownership
        )
        character_name = character_ownership.character.character_name
        messages.success(
            request, _(f"Successfully added new fleet commander: {character_name}")
        )

    return __capture_fleet_and_redirect(request, fleet_commander)


@permission_required("fleetcomp.view_all")
def capture_other_fleet_composition(request):
    """
    Enables to capture the fleet composition of another FC that registered
    """

    if request.method == "POST":
        commander_id = request.POST.get("commander_id")
        logger.info(
            "Capturing known commander fleet from commander id %s", commander_id
        )

        commander = get_object_or_404(FleetCommander, id=commander_id)

        return __capture_fleet_and_redirect(request, commander)

    logger.debug("Rendering all commander view")

    commanders = FleetCommander.objects.all()
    return render(
        request, "fleetcomp/all_fleet_commanders.html", {"commanders": commanders}
    )


def __capture_fleet_and_redirect(request, fleet_commander: FleetCommander):
    """
    Captures the fleet under this commander and returns a redirect to the snapshot if successful
    """

    logger.debug("Capturing fleet of commander %s", fleet_commander)

    try:
        fleet_id = get_fleet_id(fleet_commander)
        logger.debug("Fleet id found: %s", fleet_id)
    except CharacterNotInFleet:
        logger.error("The commander %s isn't in fleet", fleet_commander)
        messages.error(request, _("The character doesn't appear to be in a fleet"))
        return redirect("fleetcomp:index")

    if FleetSnapshot.objects.fleet_id_cache_hit(fleet_id):
        logger.info("Fleet cache hit. Returning the previous fleet")
        messages.warning(
            request,
            _(
                f"This fleet has already been captured too recently."
                f"You need to wait {FLEETCOMP_FLEET_CACHE_MINUTES} minutes between captures"
            ),
        )

        snapshot = FleetSnapshot.objects.get_last_fleet_id_record(fleet_id)

    else:
        logger.debug("No cache hit for fleet id %s", fleet_id)
        try:
            fleet_data = get_fleet_data(fleet_commander)
        except CharacterNotInFleet:
            logger.error("The commander isn't in a fleet")
            messages.error(request, _("The character doesn't appear to be in a fleet"))
            return redirect("fleetcomp:index")
        except NoAccessToFleet:
            logger.error("The character isn't the commander of its fleet")
            messages.error(
                request, _("The character doesn't appear to be a fleet commander")
            )
            return redirect("fleetcomp:index")

        snapshot = FleetSnapshot.objects.create_from_fleet_data(fleet_data)

    logger.debug(snapshot)
    return redirect("fleetcomp:view_snapshot", snapshot.id)


@permission_required("fleetcomp.basic_access")
def view_fleet(request, snapshot_id):
    """Displays the selected snapshot"""

    snapshot = get_object_or_404(FleetSnapshot, id=snapshot_id)
    logger.debug(snapshot)

    if snapshot.commander.user != request.user and not request.user.has_perm(
        "fleetcomp.view_all"
    ):
        logger.warning(
            "user %d tried to access the snapshot id %d", request.user, snapshot.id
        )
        messages.warning(
            request, _("You don't have the necessary roles to see this fleet")
        )
        return redirect("fleetcomp:index")

    main_ship_type = snapshot.get_main_ship_type()
    main_ship_type_count = snapshot.count_ship_type(main_ship_type)
    ship_groupings = ShipGrouping.objects.all()

    column_ids = (
        ShipGrouping.objects.order_by("column_index")
        .values_list("column_index", flat=True)
        .distinct()
    )
    columns = []
    for column_id in column_ids:
        columns.append(ShipGrouping.objects.filter(column_index=column_id))

    return render(
        request,
        "fleetcomp/snapshot.html",
        {
            "snapshot": snapshot,
            "main_ship_type": main_ship_type,
            "main_ship_type_count": main_ship_type_count,
            "ship_groupings": ship_groupings,
            "columns": columns,
        },
    )


@permission_required("fleetcomp.basic_access")
def user_details(request, snapshot_id: int, user_id: int | None = None):
    """Displays the details of a user"""

    snapshot = get_object_or_404(FleetSnapshot, id=snapshot_id)
    user = get_object_or_404(User, id=user_id) if user_id else None

    members = snapshot.get_user_members(user)

    context = {
        "user": user,
        "members": members,
    }

    return render(request, "fleetcomp/modals/user_details.html", context)


@cache_page(3600)
def modal_loader_body(request):
    """Draw the loader body. Useful for showing a spinner while loading a modal."""
    return render(request, "fleetcomp/modals/loader_body.html")



"""Models."""

from typing import ClassVar

from django.contrib.auth.models import User
from django.db import models
from django.db.models import QuerySet
from django.db.models.aggregates import Count
from django.utils.translation import gettext_lazy as _
from esi.models import Token
from eveuniverse.models import EveGroup, EveType

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

from . import ESI_SCOPES
from .app_settings import FLEETCOMP_DREADNOUGHT_WEIGHT_PENALTY
from .managers import (
    FleetMemberManager,
    FleetMemberQuerySet,
    FleetSnapshotManager,
    ShipGroupingManager,
)

logger = get_extension_logger(__name__)


EVE_ONLINE_DREADNOUGHT_GROUP = 485


class General(models.Model):
    """A metamodel for app permissions."""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can create snapshots and see own snapshots"),
            (
                "view_all",
                "Can view all recorded fleets and snapshot other people fleets",
            ),
        )


class ShipGrouping(models.Model):
    """Ship grouping"""

    objects: ClassVar[ShipGroupingManager] = ShipGroupingManager()

    column_index = models.PositiveIntegerField(
        default=0, help_text=_("Index of the column this grouping should be a part of")
    )
    display_name = models.CharField(
        max_length=50, help_text="Name of the grouping in the UI"
    )

    associated_types = models.ManyToManyField(
        EveType, help_text="Ship types associated to this grouping", blank=True
    )
    associated_groups = models.ManyToManyField(
        EveGroup, help_text="Eve ship groups associated to this grouping", blank=True
    )

    @property
    def internal_name(self) -> str:
        """Internal name to use when passing value around"""
        return self.display_name.lower().replace(" ", "_")

    def __str__(self):
        return f"Custom group for {self.display_name}"

    def get_snapshot_matches(
        self, snapshot: "FleetSnapshot"
    ) -> QuerySet["FleetMember"]:
        """Return all members matching the grouping"""
        return snapshot.members.all().filter_by_ship_grouping(self)


class FleetCommander(models.Model):
    """
    Member of a fleet with ESI stored
    """

    character_ownership = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.CASCADE,
        related_name="+",
    )

    @property
    def character_id(self) -> int:
        """Return character id"""
        return self.character_ownership.character.character_id

    @property
    def character_name(self) -> str:
        """Return character name"""
        return self.character_ownership.character.character_name

    @property
    def user(self) -> User:
        """Return the associated user"""
        return self.character_ownership.user

    def fetch_token(self) -> Token:
        """Return a valid token if there is one"""
        token = (
            Token.objects.filter(
                character_id=self.character_id,
            )
            .require_scopes(ESI_SCOPES)
            .require_valid()
            .first()
        )

        if not token:
            raise Token.DoesNotExist(f"{self}: No valid token found")
        return token

    def __str__(self):
        return self.character_name


class FleetSnapshot(models.Model):
    """Takes a snapshot of a fleet at a given time"""

    objects: ClassVar[FleetSnapshotManager] = FleetSnapshotManager()

    fleet_id = models.BigIntegerField(help_text=_("EVE online fleet id"))
    timestamp = models.DateTimeField(
        auto_now_add=True, help_text=_("Time of the snapshot")
    )

    commander = models.ForeignKey(
        FleetCommander,
        on_delete=models.SET_NULL,
        null=True,
        help_text=_("Fleet commander at the time of the snapshot"),
    )

    @property
    def timestamp_str(self) -> str:
        """Timestamp displayed with leading zeroes"""
        return self.timestamp.strftime("%Y/%m/%d %H:%M")

    class Meta:
        ordering = ["-timestamp"]

    def __str__(self):
        return f"Fleet {self.fleet_id} at {self.timestamp}"

    def count_mains(self) -> int:
        """Counts how many main characters are in the fleet"""
        return (
            self.members.exclude(character__character_ownership=None)
            .values("character__character_ownership__user")
            .distinct()
            .count()
        )

    def get_user_ids(self):
        """Return the list of users in this fleet"""
        return self.members.values_list(
            "character__character_ownership__user", flat=True
        ).distinct()

    def count_orphans(self) -> int:
        """Count how many characters in the fleet don't have a main associated"""
        return self.members.filter(character__character_ownership=None).count()

    def count_members(self) -> int:
        """Counts how many characters are in the fleet"""
        return self.members.count()

    def get_user_members(self, user: User | None) -> FleetMemberQuerySet:
        """
        Returns characters in fleets linked to this user.
        If None is passed returns all orphans
        """
        if user:
            user_members = self.members.filter(
                character__character_ownership__user=user
            )
        else:
            user_members = self.members.filter(character__character_ownership=None)
        return user_members

    def get_main_ship_type(self) -> EveType:
        """Return the most popular ship type of this fleet"""

        def reduce_dreadnought_weight(ship_type_count):
            """
            Utility function reducing the weight of dreadnoughts to have them removed from main ship
            """
            if ship_type_count["ship_type__eve_group"] == EVE_ONLINE_DREADNOUGHT_GROUP:
                ship_type_count["id__count"] /= FLEETCOMP_DREADNOUGHT_WEIGHT_PENALTY
            return ship_type_count

        ship_type_counts = self.members.values(
            "ship_type", "ship_type__eve_group"
        ).annotate(Count("id"))
        penalized_ship_type_counts = list(
            map(reduce_dreadnought_weight, ship_type_counts)
        )
        print(penalized_ship_type_counts)
        penalized_ship_type_counts.sort(
            key=lambda ship_type_count: ship_type_count["id__count"], reverse=True
        )

        return EveType.objects.get(id=penalized_ship_type_counts[0]["ship_type"])

    def count_ship_type(self, ship_type: EveType) -> int:
        """Returns how many of a certain ship there is in the fleet"""
        return self.members.filter(ship_type=ship_type).count()

    def get_user_and_associated_members_list(
        self,
    ) -> list[tuple[User | None, list["FleetMember"]]]:
        """Returns all users and their associated member list"""
        res = []
        for user_id in self.get_user_ids():
            if user_id:
                user = User.objects.get(id=user_id)
                res.append((user, self.get_user_members(user)))
            else:
                res.append((None, self.get_user_members(None)))

        return res

    def get_total_mass(self) -> float:
        """Return the total ship mass in tons"""
        return self.members.all().calculate_mass() / 1000

    def get_subcap_mass(self) -> float:
        """Return the mass of subcap ships in tons"""
        return self.members.all().exclude_capitals().calculate_mass() / 1000


class FleetMember(models.Model):
    """Member of a fleet and its associated ships"""

    objects: ClassVar[FleetMemberManager] = FleetMemberManager()

    fleet = models.ForeignKey(
        FleetSnapshot, on_delete=models.CASCADE, related_name="members"
    )
    character = models.ForeignKey(EveCharacter, on_delete=models.CASCADE)
    ship_type = models.ForeignKey(EveType, on_delete=models.CASCADE)

    def __str__(self):
        return f"{self.character} - {self.ship_type}"



"""Fleetcomp templatetags"""

from django import template
from django.contrib.auth.models import User
from eveuniverse.models import EveType

from fleetcomp.managers import FleetMemberQuerySet
from fleetcomp.models import FleetSnapshot, ShipGrouping

register = template.Library()


@register.simple_tag
def custom_grouping_counter_snapshot(
    snapshot: FleetSnapshot, custom_grouping: ShipGrouping
) -> int:
    """Returns how many ships of the given custom grouping are in the snapshot"""
    return custom_grouping.get_snapshot_matches(snapshot).count()


@register.simple_tag
def custom_grouping_counter_user(
    snapshot: FleetSnapshot, user: User, custom_grouping: ShipGrouping
) -> int:
    """Returns how many ships of the given user match the custom grouping"""
    return (
        custom_grouping.get_snapshot_matches(snapshot) & snapshot.get_user_members(user)
    ).count()


@register.simple_tag
def count_ship_type_in_fleet_members(
    ship_type: EveType, fleet_members: FleetMemberQuerySet
):
    """Count how many fleet members are in the given ship type"""
    return fleet_members.count_ship_type(ship_type)


@register.simple_tag
def count_ships_in_grouping(
    ship_grouping: ShipGrouping, fleet_members: FleetMemberQuerySet
):
    """Count the number of members with a ship matching this grouping"""
    return fleet_members.filter_by_ship_grouping(ship_grouping).count()


@register.simple_tag
def count_ships_not_in_groupings_or_mainline(
    fleet_members: FleetMemberQuerySet, mainline: EveType
):
    """Count members with a ship not matching any grouping"""
    return (
        fleet_members.exclude_ships_in_any_grouping()
        .exclude(ship_type=mainline)
        .count()
    )



# Generated by Django 4.2.21 on 2025-06-25 19:33
# Additional template models added by manually by T'rahk

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
        ("fleetcomp", "0001_initial"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="CustomGrouping",
            new_name="ShipGrouping",
        ),
        migrations.AddField(
            model_name="shipgrouping",
            name="column_index",
            field=models.PositiveIntegerField(
                default=0,
                help_text="Index of the column this grouping should be a part of",
            ),
        ),
    ]



"""Routes."""

from django.urls import path

from . import views

app_name = "fleetcomp"

urlpatterns = [
    path("", views.index, name="index"),
    path("fleet/add/own", views.capture_own_fleet_composition, name="add_own_snapshot"),
    path(
        "fleet/add/other",
        views.capture_other_fleet_composition,
        name="add_other_snapshot",
    ),
    path("fleet/<int:snapshot_id>", views.view_fleet, name="view_snapshot"),
    path(
        "fleet/<int:snapshot_id>/details",
        views.user_details,
        name="user_details_orphans",
    ),
    path(
        "fleet/<int:snapshot_id>/details/<int:user_id>",
        views.user_details,
        name="user_details",
    ),
    path("modal_loader_body", views.modal_loader_body, name="modal_loader_body"),
]



from django.test import TestCase
from eveuniverse.tools.testdata import ModelSpec, create_testdata

from . import test_data_filename


class CreateEveuniverseTestData(TestCase):
    def test_create_testdata(self):
        test_data_spec = [
            ModelSpec(
                "EveType",
                ids=[
                    28352,  # Rorqual
                    19720,  # Revelation
                    11567,  # Avatar
                    621,  # Caracal
                    585,  # Slasher
                ],
            )
        ]
        create_testdata(test_data_spec, test_data_filename())



"""App settings."""

from django.conf import settings

FLEETCOMP_FLEET_CACHE_MINUTES = getattr(settings, "FLEETCOMP_FLEET_CACHE_MINUTES", 5)

FLEETCOMP_DREADNOUGHT_WEIGHT_PENALTY = getattr(
    settings, "FLEETCOMP_DREADNOUGHT_WEIGHT_PENALTY", 5
)



"""Managers."""

from datetime import timedelta
from enum import IntEnum

from django.db import models
from django.db.models import Q, Sum
from django.utils.timezone import now
from eveuniverse.models import EveType

from allianceauth.eveonline.models import EveCharacter

from . import esi
from . import models as fleetcomp_models
from .app_settings import FLEETCOMP_FLEET_CACHE_MINUTES


class EveGroupIds(IntEnum):
    """Groups id of several Eve groups"""

    FORCE_AUXILIARY = 1538
    DREADNOUGHT = 485
    CARRIER = 547
    SUPERCARRIER = 659
    TITAN = 30
    FREIGHTER = 513
    CAPITAL_INDUSTRIAL_SHIP = 883


class ShipGroupingManager(models.Manager):
    """Manager for ShipGrouping"""

    def get_all_type_ids(self) -> list[int]:
        """Return all type ids present in groupings"""
        return list(self.values_list("associated_types__id", flat=True).distinct())

    def get_all_group_ids(self) -> list[int]:
        """Return all group ids present in groupings"""
        return list(self.values_list("associated_groups__id", flat=True).distinct())


class FleetSnapshotManager(models.Manager):
    """Manager for FLeetSnapshot"""

    def create_from_fleet_data(
        self, fleet_data: "esi.FleetData"
    ) -> "fleetcomp_models.FleetSnapshot":
        """Create a snapshot from the data received by the ESI"""

        snapshot: "fleetcomp_models.FleetSnapshot" = self.create(
            fleet_id=fleet_data.fleet_id,
            commander=fleet_data.fleet_commander,
        )

        for member in fleet_data.fleet_members:
            try:
                character = EveCharacter.objects.get(character_id=member.character_id)
            except EveCharacter.DoesNotExist:
                character = EveCharacter.objects.create_character(
                    character_id=member.character_id
                )
            ship_type, _ = EveType.objects.get_or_create_esi(id=member.ship_type_id)
            fleetcomp_models.FleetMember.objects.create(
                character=character, ship_type=ship_type, fleet=snapshot
            )

        return snapshot

    def fleet_id_cache_hit(self, fleet_id: int) -> bool:
        """
        Returns true if there is a fleet in the database created less than FLEETCOMP_FLEET_CACHE_MINUTES minutes ago
        """
        return self.filter(
            fleet_id=fleet_id,
            timestamp__gt=now() + timedelta(minutes=-FLEETCOMP_FLEET_CACHE_MINUTES),
        ).exists()

    def get_last_fleet_id_record(self, fleet_id) -> "fleetcomp_models.FleetSnapshot":
        """Returns the last record of a fleet with this id"""
        # TODO check that the ordering is correct
        return self.filter(fleet_id=fleet_id).first()


class FleetMemberQuerySet(models.QuerySet):
    """Custom QuerySet class for FleetMember"""

    def filter_by_ship_grouping(
        self, ship_grouping: "fleetcomp_models.ShipGrouping"
    ) -> models.QuerySet:
        """Filters fleet members if their ship type matches the given ship grouping"""
        return self.filter(
            Q(ship_type__eve_group__in=ship_grouping.associated_groups.all())
            | Q(ship_type__in=ship_grouping.associated_types.all())
        )

    def exclude_ships_in_any_grouping(self):
        """Exccludes all fleet members with a ship that would match an existing queryset"""
        excluded_groups_ids = fleetcomp_models.ShipGrouping.objects.get_all_group_ids()
        excluded_types_ids = fleetcomp_models.ShipGrouping.objects.get_all_type_ids()

        return self.exclude(ship_type__eve_group__id__in=excluded_groups_ids).exclude(
            ship_type__id__in=excluded_types_ids
        )

    def filter_by_ship_type(self, ship_type: EveType):
        """Filters by the members ship type"""
        return self.filter(ship_type=ship_type)

    def count_ship_type(self, ship_type: EveType) -> int:
        """Count how many fleet members have the given ship type"""
        return self.filter_by_ship_type(ship_type).count()

    def exclude_capitals(self):
        """Will exclude all members sitting in a capital ship"""
        return self.exclude(
            ship_type__eve_group__id__in=[
                EveGroupIds.FORCE_AUXILIARY,
                EveGroupIds.DREADNOUGHT,
                EveGroupIds.CARRIER,
                EveGroupIds.FREIGHTER,
                EveGroupIds.SUPERCARRIER,
                EveGroupIds.TITAN,
                EveGroupIds.CAPITAL_INDUSTRIAL_SHIP,
            ]
        )

    def calculate_mass(self) -> float:
        """Returns the sum of the mass of all ships in this queryset"""
        return self.aggregate(mass_total=Sum("ship_type__mass"))["mass_total"]


class FleetMemberManager(models.Manager):
    """Custom Manager for FleetMember"""

    def get_queryset(self) -> FleetMemberQuerySet:
        """Custom get_queryset method"""
        return FleetMemberQuerySet(self.model, using=self._db)



# Generated by Django 4.2.21 on 2025-06-14 18:39

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("authentication", "0024_alter_userprofile_language"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can create snapshots and see own snapshots"),
                    (
                        "view_all",
                        "Can view all recorded fleets and snapshot other people fleets",
                    ),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="FleetCommander",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "character_ownership",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="authentication.characterownership",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FleetSnapshot",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("fleet_id", models.BigIntegerField(help_text="EVE online fleet id")),
                (
                    "timestamp",
                    models.DateTimeField(
                        auto_now_add=True, help_text="Time of the snapshot"
                    ),
                ),
                (
                    "commander",
                    models.ForeignKey(
                        help_text="Fleet commander at the time of the snapshot",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="fleetcomp.fleetcommander",
                    ),
                ),
            ],
            options={
                "ordering": ["-timestamp"],
            },
        ),
        migrations.CreateModel(
            name="FleetMember",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveonline.evecharacter",
                    ),
                ),
                (
                    "fleet",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="members",
                        to="fleetcomp.fleetsnapshot",
                    ),
                ),
                (
                    "ship_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="CustomGrouping",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "display_name",
                    models.CharField(
                        help_text="Name of the grouping in the UI", max_length=50
                    ),
                ),
                (
                    "associated_groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Eve ship groups associated to this grouping",
                        to="eveuniverse.evegroup",
                    ),
                ),
                (
                    "associated_types",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Ship types associated to this grouping",
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
        ),
    ]



from django.test import TestCase
from eveuniverse.models import EveType

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.tests.auth_utils import AuthUtils

from fleetcomp.models import FleetCommander, FleetMember, FleetSnapshot
from fleetcomp.tests.testdata.load_eveuniverse import load_eveuniverse


class TestSnapshot(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        # create a global fleet commander
        cls.user = AuthUtils.create_user("test_user")
        cls.character = AuthUtils.add_main_character_2(
            cls.user, "test character", 1, "123", "Test corporation", "TEST"
        )
        character_ownership = CharacterOwnership.objects.create(
            character=cls.character,
            user=cls.user,
            owner_hash="fake hash",
        )
        cls.fleet_commander = FleetCommander.objects.create(
            character_ownership=character_ownership,
        )

    def test_main_ship_reduce_dread(self):
        fleet_snapshot: FleetSnapshot = FleetSnapshot.objects.create(
            fleet_id=10,
            commander=self.fleet_commander,
        )

        revelation_type = EveType.objects.get(id=19720)
        caracal_type = EveType.objects.get(id=621)
        FleetMember.objects.create(
            fleet=fleet_snapshot, character=self.character, ship_type=revelation_type
        )
        FleetMember.objects.create(
            fleet=fleet_snapshot, character=self.character, ship_type=revelation_type
        )
        FleetMember.objects.create(
            fleet=fleet_snapshot, character=self.character, ship_type=caracal_type
        )

        self.assertEqual(fleet_snapshot.get_main_ship_type(), caracal_type)



from django.test import TestCase
from eveuniverse.models import EveType

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.tests.auth_utils import AuthUtils

from ..models import FleetCommander, FleetMember, FleetSnapshot
from .testdata.load_eveuniverse import load_eveuniverse


class TestFleetMemberQuerySet(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.test_user = AuthUtils.create_user("test_user")
        AuthUtils.add_main_character_2(
            user=cls.test_user, name="test_char", character_id=1234
        )
        char_ownership = CharacterOwnership.objects.create(
            user=cls.test_user,
            character=cls.test_user.profile.main_character,
            owner_hash="badhash",
        )

        cls.fleet_commander = FleetCommander.objects.create(
            character_ownership=char_ownership
        )

    def setUp(self):
        super().setUp()
        self.snapshot = FleetSnapshot.objects.create(
            fleet_id=123, commander=self.fleet_commander
        )

    def test_exclude_capitals(self):
        FleetMember.objects.bulk_create(
            [
                FleetMember(
                    fleet=self.snapshot,
                    character=self.test_user.profile.main_character,
                    ship_type=EveType.objects.get(id=621),
                ),
                FleetMember(
                    fleet=self.snapshot,
                    character=self.test_user.profile.main_character,
                    ship_type=EveType.objects.get(id=11567),
                ),
                FleetMember(
                    fleet=self.snapshot,
                    character=self.test_user.profile.main_character,
                    ship_type=EveType.objects.get(id=28352),
                ),
            ]
        )
        qs = FleetMember.objects.all()

        self.assertEqual(qs.count(), 3)

        qs = qs.exclude_capitals()

        self.assertEqual(qs.count(), 1)

    def test_total_mass(self):
        FleetMember.objects.bulk_create(
            [
                FleetMember(
                    fleet=self.snapshot,
                    character=self.test_user.profile.main_character,
                    ship_type=EveType.objects.get(id=621),
                ),
                FleetMember(
                    fleet=self.snapshot,
                    character=self.test_user.profile.main_character,
                    ship_type=EveType.objects.get(id=11567),
                ),
            ]
        )

        qs = FleetMember.objects.all()

        self.assertEqual(qs.calculate_mass(), 11910000.0 + 2400000000.0)
